from __future__ import annotations
import traceback

from inspect import iscoroutinefunction
import asyncio
import logging

import js
from js import console

import wwwpy.common.reloader as reloader
from wwwpy.common import _no_remote_infrastructure_found_text, files, _remote_module_not_found_html
from wwwpy.common.tree import print_tree
from wwwpy.remote.designer import dev_mode as dm
from wwwpy.remote.websocket import setup_websocket
from wwwpy.common.asynclib import create_task_safe

logger = logging.getLogger(__name__)


async def entry_point(dev_mode: bool = False):
    # from wwwpy.common.tree import print_tree
    # print_tree('/wwwpy_bundle')

    await setup_websocket()
    if dev_mode:
        await dm.activate()

    await _invoke_browser_main()


def _reload():
    async def reload():
        console.log('reloading')
        reloader.unload_path(files._bundle_path)
        await _invoke_browser_main()

    asyncio.create_task(reload())


async def _invoke_browser_main():
    try:
        console.log('invoke_browser_main')

        try:
            js.document.body.innerText = f'Going to import the "remote" package'
            import remote
            if hasattr(remote, 'main'):
                if iscoroutinefunction(remote.main):
                    await remote.main()
                else:
                    remote.main()
        except ModuleNotFoundError as e:
            js.document.body.innerHTML = _remote_module_not_found_html
        except Exception as e:
            _show_exception(e, _no_remote_infrastructure_found_text)
            from wwwpy.server.designer import rpc
            create_task_safe(rpc.on_exception_string(traceback.format_exc()))
    finally:
        if dm.is_active():
            from wwwpy.remote.designer.ui import dev_mode_component
            dev_mode_component.show()


def _show_exception(e, html: str):
    js.document.body.innerHTML = html
    js.document.body.insertAdjacentHTML('beforeend', '<br><br>')

    exc_trace = traceback.format_exc()
    exc_str = 'Exception: ' + str(e) + '\n\n' + exc_trace + '\n\n'
    pre = js.document.createElement('pre')
    pre.innerText = exc_str
    js.document.body.appendChild(pre)



from __future__ import annotations

import threading
from datetime import timedelta
from threading import Thread, Event
from time import sleep
from typing import Callable, Any, List

from wwwpy.server.filesystem_sync.debouncer import Debouncer


class DebouncerThread:

    def __init__(self, debouncer: Debouncer, emit: Callable[[List[Any]], None]):
        self._debouncer = debouncer
        self._emit = emit
        self._event = Event()
        debouncer.wakeup = self._wakeup
        self._continue = True
        self._thread: Thread | None = None

    def _wakeup(self, debouncer: Debouncer | None):
        self._event.set()

    def _thread_loop(self):
        while self._continue:
            delta = self._debouncer.time_until_next_emission()
            timeout_millis = 1 if delta is None else delta.total_seconds()
            self._event.wait(timeout_millis)
            self._event.clear()
            events = self._debouncer.events()
            if events:
                self._emit(events)
        self._thread = None

    def start(self):
        if self._thread is not None:
            raise RuntimeError('Thread already started')
        self._continue = True
        self._thread = Thread(daemon=True, target=self._thread_loop, name='DebouncerThread')
        self._thread.start()

    def stop(self):
        self._continue = False
        self._wakeup(None)

    def join(self):
        t = self._thread
        if t is not None:
            t.join()

def main():
    debouncer = Debouncer(timedelta(milliseconds=100))

    def print_events(events):
        print('=' * 20 + f' current thread name={threading.current_thread().name}')
        print(f'len={len(events)} events={events}')

    debouncer_thread = DebouncerThread(debouncer, print_events)
    counter = 0

    def next_event():
        nonlocal counter
        counter += 1
        return f"e{counter}"

    def send(n, between, final):
        for _ in range(n, ):
            debouncer.add_event(next_event())
            sleep(between)
        sleep(final)

    for t in range(10):
        send(5, 0.09, 1)
        send(100, 0.01, 0.3)
        send(10, 0.05, 0.5)

    debouncer_thread.stop_join()


if __name__ == '__main__':
    main()



import logging
import sys
from pathlib import Path

from wwwpy.common import modlib
from wwwpy.common.designer import new_component

logger = logging.getLogger(__name__)


async def write_module_file(module: str, content: str) -> str:
    msg = f'write_module_file module={module} content len={len(content)}'
    logger.debug(msg)
    path = modlib._find_module_path(module)
    if not path:
        raise ValueError(f'Cannot find module {module}')
    path.write_text(content)
    return f'done {msg}'


def _fix_stacktrace(message: str):
    return message.replace('"/wwwpy_bundle/', '"')

async def on_exception_string(exception: str):
    exception = _fix_stacktrace(exception)
    print(exception, file=sys.stderr)

async def on_error(message: str, source: str, lineno: int, colno: int, error: str):
    print(f'rpc.on_error')
    message = _fix_stacktrace(message)
    error = _fix_stacktrace(error)
    all_str = f'message==={message}\nsource==={source}\nlineno==={lineno}\ncolno==={colno}\nerror==={error}'
    print(all_str, file=sys.stderr)


async def on_unhandledrejection(message: str):
    message = _fix_stacktrace(message)
    print(message, file=sys.stderr)


async def print_module_line(module: str, message: str, lineno: int):
    #  File "remote/designer/toolbar.py", line 140, in on_changes
    path = modlib._find_module_path(module)
    if not path:
        raise ValueError(f'Cannot find module {module}')
    print(message)
    print(_ide_path_link(path, lineno))


def _ide_path_link(path, lineno):
    ide_path_link = f'  File "{path}", line {lineno}'
    return ide_path_link


async def server_console_log(message: str):
    print(message)


async def add_new_component() -> str:
    path = new_component.add()
    if path:
        print(f'Added new component')
        print(_ide_path_link(path, 1))
        return f'Added new component file:\n\n{path.name}'
    return 'Failed to add new component'


async def quickstart_apply(quickstart_name: str) -> str:
    if not await quickstart_possible():
        return 'Quickstart not possible'
    from wwwpy.server import custom_str
    root = custom_str.get_root_folder_or_fail()
    from wwwpy.common import quickstart
    quickstart.setup_quickstart(Path(root), quickstart_name)
    logger.info(f'Quickstart applied {quickstart_name} to {root}')
    return 'Quickstart applied'


async def quickstart_possible() -> bool:
    from wwwpy.server import custom_str
    from wwwpy.common import quickstart
    root = custom_str.get_root_folder_or_fail()
    emtpy = quickstart.is_empty_project(root)
    return emtpy



from __future__ import annotations

import os
import argparse
from pathlib import Path
from typing import NamedTuple, Optional, Sequence

from wwwpy.server.settingslib import user_settings


class Arguments(NamedTuple):
    directory: str
    port: int
    dev: bool


def parse_arguments(args: Optional[Sequence[str]] = None) -> Arguments:
    parser = argparse.ArgumentParser(prog='wwwpy')
    parser.add_argument('dev', nargs='?', const=True, default=False,
                        help="Run in development mode")
    parser.add_argument('--directory', '-d', default=os.getcwd(),
                        help='set the root path for the project (default: current directory)')
    parser.add_argument('--port', type=int, default=8000,
                        help='bind to this port (default: 8000)')

    parsed_args = parser.parse_args(args)
    return Arguments(
        directory=parsed_args.directory,
        port=parsed_args.port,
        dev=bool(parsed_args.dev)
    )


def run_server(args: Arguments):
    import wwwpy
    print(f'Starting wwwpy v{wwwpy.__version__}')
    from wwwpy.server import configure
    from wwwpy.webserver import wait_forever

    working_dir = Path(args.directory).absolute()
    settings = user_settings()
    configure.start_default(args.port, working_dir, dev_mode=args.dev, settings=settings)
    _open_browser(args, settings)

    wait_forever()


def _open_browser(args, settings):
    import webbrowser
    if settings.open_url_code:
        loc = {'url': f'http://localhost:{args.port}', args: args}
        exec(settings.open_url_code, loc, loc)
    else:
        webbrowser.open(f'http://localhost:{args.port}')


def main():
    args = parse_arguments()
    run_server(args)


if __name__ == '__main__':
    main()



import wwwpy.remote.component as wpc
import js

import logging
import urllib.parse

logger = logging.getLogger(__name__)


class MailtoComponent(wpc.Component):
    recipient: str = wpc.attribute()
    subject: str = wpc.attribute()
    body: str = wpc.attribute()
    text_content: str = wpc.attribute()
    target: js.HTMLElement = wpc.element()

    _link: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """<a data-name="_link"></a>"""

        self._update_attributes()

    def attributeChangedCallback(self, name: str, oldValue: str, newValue: str):
        self._update_attributes()

    def connectedCallback(self):
        self._update_attributes()

    def _update_attributes(self):
        target = self.target if self.target else '_blank'
        recipient = self.recipient if self.recipient else 'foo@example.com'
        subject = self.subject if self.subject else 'subject not specified'
        body = self.body if self.body else ''
        text_content = self.text_content if self.text_content else f"Send Email to {recipient}"

        # make sure to encode the subject and body
        subject = urllib.parse.quote_plus(subject)
        body = urllib.parse.quote_plus(body)

        self._link.href = f"mailto:{recipient}?subject={subject}&body={body}"
        self._link.textContent = text_content
        self._link.target = target



import asyncio

import wwwpy.remote.component as wpc
import js

import logging

logger = logging.getLogger(__name__)


class Component1(wpc.Component, tag_name='component-1'):
    textarea1: js.HTMLTextAreaElement = wpc.element()
    div1: js.HTMLDivElement = wpc.element()
    title: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div data-name="title">Loading...</div>

<style>
.textarea1 {
    width: 100%;
    padding: 10px;
    border: 1px solid #ccc;
    border-radius: 5px;
    font-family: Arial, sans-serif;
    font-size: 14px;
    resize: vertical;
    box-shadow: 0 2px 5px rgba(0, 0, 0, 0.1);
}
</style>

<textarea data-name="textarea1" rows="7" class="textarea1"></textarea>
<hr>
<div data-name="div1"></div>

"""
        asyncio.ensure_future(self.after_init())

    async def after_init(self):
        logger.info('Component1 after_init')
        import micropip
        await micropip.install('markdown')
        import markdown
        # language=markdown
        self.title.innerHTML = markdown.markdown(f"""
# Component1
 - This is a simple component that uses a textarea and a div to render markdown.
 - This text itself is rendered using markdown.      
""")
        self.textarea1.value = js.localStorage.getItem('textarea1') or 'Write your _markdown_ here ...'
        self._render_markdown()

    async def textarea1__input(self, event):
        js.localStorage.setItem('textarea1', self.textarea1.value)
        self._render_markdown()

    def _render_markdown(self):
        import markdown
        self.div1.innerHTML = markdown.markdown(self.textarea1.value)



import dataclasses
import shutil
from pathlib import Path
from typing import List, Dict, Iterable, Union, Optional

from wwwpy.common import tree
from wwwpy.common.filesystem.sync import Event
from dataclasses import dataclass, field


def events_invert(fs: Path, events: List[Event]) -> List[Event]:
    root = Node('.', '.', True)

    """This is the root node of the tree that will be used to keep track of the path changes"""

    def _get_or_create_node(path_str: str) -> Node:
        path_node = _get_node_chain(root, path_str)[-1]
        if path_node is None:
            is_dir = (fs / path_str).is_dir()
            path_node = _create_node(root, path_str, is_dir)
        return path_node

    def get_final_path(event: Event) -> Path:
        path_node = _get_or_create_node(event.src_path)
        assert path_node is not None
        return Path(fs / path_node.final_path)

    def augment(event: Event) -> Event:
        fp = get_final_path(event)
        content = _get_content(fp)
        aug = dataclasses.replace(event, content=content)
        return aug

    def is_deleted_entity(rel: Event) -> bool:
        chain = _get_node_chain(root, rel.src_path)
        for node in chain:
            if node is not None and node.is_deleted:
                return True
        return False

    def is_modified_entity(rel: Event) -> bool:
        node = _get_node_chain(root, rel.src_path)[-1]
        if node is not None and node.is_modified:
            return True
        return False

    relative_events = []
    for e in reversed(events):
        if e.event_type == 'closed':
            continue
        rel = e.relative_to(fs)
        if rel.src_path == '' or rel.src_path == '.':
            continue  # skip any event on the root
        if is_deleted_entity(rel):
            continue
        # we are processing the events backwards in time to go from A_n to A_0
        if e.event_type == 'moved':
            current_path = rel.dest_path  # this is the path in A_i
            new_path = rel.src_path  # this is the path in A_(i-1)
            final_node = _get_node_chain(root, current_path)[-1]
            if final_node is None:
                # this is the first rename of this path
                _create_node(root, current_path, (fs / current_path).is_dir())

            _move_node(root, current_path, new_path)  # we mutate the tree backwards in time

        if is_modified_entity(rel):
            continue
        if rel.event_type == 'deleted':
            # mark this entity such as we are ignore all preceding events, 'moved' included
            _get_or_create_node(rel.src_path).is_deleted = True
        elif rel.event_type == 'modified':
            if get_final_path(rel).is_dir():
                continue
            rel = augment(rel)
            # mark this entity such as we are ignore all preceding events, except 'moved'
            _get_or_create_node(rel.src_path).is_modified = True

        relative_events.insert(0, rel)

    return relative_events


def events_apply(fs: Path, events: List[Event]):
    for event in events:
        _event_apply(fs, event)


def _event_apply(fs: Path, event: Event):
    path = fs / event.src_path
    t = event.event_type
    is_dir = event.is_directory
    if t == 'created':
        if is_dir:
            path.mkdir(exist_ok=True)  # todo this is because of dev_mode on rpc signature change
        else:
            path.touch()
    elif t == 'deleted':
        if path.exists():  # it could not exist because of events compression
            if is_dir:
                shutil.rmtree(path)  # again because of events compression we could need to remove a whole tree
            else:
                path.unlink()
    elif t == 'moved':
        shutil.move(path, fs / event.dest_path)
    elif t == 'modified':
        c = event.content
        if isinstance(c, str):
            path.write_text(c)
        elif isinstance(c, bytes):
            path.write_bytes(c)
        else:
            raise ValueError(f"Unsupported content type: {type(c)}")


def _get_content(path: Path):
    assert path.exists(), f'Path does not exist: {path}'
    assert path.is_file(), f'Not a file: {path}'
    try:
        return path.read_text()
    except UnicodeDecodeError:
        return path.read_bytes()


@dataclass
class Node:
    name: str
    """This is the instant name of the node. It is the name of the node in the A_i state"""
    final_path: str
    """This is the final path of the node in the A_n state. It is a relative complete path"""
    is_directory: bool
    is_deleted: bool = False
    is_modified: bool = False
    children: Dict[str, 'Node'] = field(default_factory=dict)

    # validate in post init
    def __post_init__(self):
        if '//' in self.final_path:
            raise ValueError(f'Invalid path: {self.final_path}')

    def str(self):
        return f'{self.name}'


def _get_node_chain(root: Node, path: str) -> List[Optional[Node]]:
    if path == '.':
        return [root]
    parts = _get_parts(path)
    current = root
    chain = [root]

    for part in parts:
        if current is None or part not in current.children:
            chain.append(None)
            current = None
        else:
            current = current.children[part]
            chain.append(current)

    return chain


def _get_parts(path):
    return path.strip("/").split("/")


def _create_node(root: Node, final_path: str, is_dir: bool) -> Node:
    if final_path == '':
        return root
    parts = _get_parts(final_path)
    current = root

    for i, name in enumerate(parts):
        if name not in current.children:
            path = current.final_path + '/' + name
            current.children[name] = Node(name, path, is_dir)
        current = current.children[name]

    return current


def _move_node(root: Node, current_path: str, new_path: str):
    """This function moves a node from the current path to the new path.
    if the new path already exists, we will raise an error"""
    print('\nbefore move')
    print_node(root)
    current_node_chain = _get_node_chain(root, current_path)
    new_node_chain = _get_node_chain(root, new_path)

    if new_node_chain[-1] is not None:
        raise ValueError(f'Path already exists: {new_path}')

    current_node = current_node_chain[-1]
    current_node_parent = current_node_chain[-2]
    new_node_parent = new_node_chain[-2]
    new_name = _get_parts(new_path)[-1]

    current_node_parent.children.pop(current_node.name)
    current_node.name = new_name
    new_node_parent.children[new_name] = current_node

    print('after move')
    print_node(root)
    return


@dataclass
class _NodePrint(tree.NodeProtocol):
    node: Node

    def iterdir(self) -> Iterable[tree.NodeProtocol]:
        return [_NodePrint(child) for child in self.node.children.values()]

    def is_dir(self) -> bool:
        return self.node.is_directory

    @property
    def name(self) -> str:
        msg = f' ({self.node.final_path}) is_dir={self.node.is_directory} is_deleted={self.node.is_deleted}'
        return self.node.str().strip('/') + msg


def print_node(node: Node, printer=print):
    for line in tree.tree(_NodePrint(node)):
        printer(line)


def events_init(source: Path) -> List[Event]:
    result = []
    for path in source.rglob('*'):
        src_path = str(path.relative_to(source))
        if path.is_dir():
            result.append(Event(event_type='created', src_path=src_path, is_directory=True))
        else:
            result.append(Event(event_type='modified', src_path=src_path, is_directory=False,
                                content=_get_content(path)))
    return result



import js
import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js

from common import name

import logging

logger = logging.getLogger(__name__)

class Component1(wpc.Component, tag_name='component-1'):
    slInput1: js.HTMLElement = wpc.element()
    slTextarea1: js.HTMLElement = wpc.element()
    slSwitch1: js.HTMLElement = wpc.element()
    nickname: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<style> * { margin-bottom: 0.5em; } </style>
<h2>You are: <span data-name="nickname"></span></h2>
<sl-switch data-name="slSwitch1" checked>Allow empty messages</sl-switch>
<sl-input data-name="slInput1" placeholder="Message"></sl-input>
<sl-button data-name="slButton1">Send<sl-icon slot="suffix" name="send"></sl-icon></sl-button>
<sl-textarea data-name="slTextarea1" placeholder="slTextarea1" rows="10">Messages...</sl-textarea>
"""
        local_storage_key = 'wwwpy.userspace.chat.nickname'
        nick = js.window.localStorage.getItem(local_storage_key)
        if not nick:
            nick = name.generate_name()
            js.window.localStorage.setItem(local_storage_key, nick)
        self.nickname.textContent = nick

        def new_message(message):
            self.slTextarea1.value += message + '\n'
            self.slTextarea1.scrollPosition(dict_to_js({'top': 999999999}))

        import remote.rpc
        remote.rpc.messages_listeners.append(new_message)

    async def slButton1__click(self, event):
        logger.debug('handler slButton1__click')
        self._set_error('')
        try:
            message = self.slInput1.value.strip()
            if not message:
                if not self.slSwitch1.checked:
                    self._set_error('Please enter a message')
                    return
                message = 'Empty message'

            import server.rpc
            await server.rpc.send_message_to_all(self.nickname.innerText + ': ' + message)
            self.slInput1.value = ''
            self._set_error('')
        except Exception as e:
            self._set_error('Issues sending message... ' + str(e))

    def _set_error(self, msg):
        self.slInput1.setCustomValidity(msg)
        self.slInput1.reportValidity()



from __future__ import annotations

from dataclasses import dataclass
from typing import Protocol, Iterator, TypeVar, Generic, Type

from wwwpy.common.rpc import serialization

import logging

logger = logging.getLogger(__name__)


def _restore(cls: Type[T], storage: Storage = None) -> T:
    if storage is None:
        storage = JsStorage()
    name = cls.__module__ + '.' + cls.__name__
    state_manager = State(storage, f'wwwpy.persist.{name}')
    result = state_manager.restore(cls).instance_or_default()

    def save():
        logger.debug(f'save {name} state res={result}')
        state_manager.save(result)

    from .property_monitor import monitor_changes
    monitor_changes(result, lambda *a: save())
    return result


class Storage(Protocol):
    @property
    def length(self) -> int: ...

    def keys(self) -> Iterator[str]: ...

    def items(self) -> Iterator[tuple[str, str]]: ...

    def getItem(self, key: str) -> str | None: ...

    def setItem(self, key: str, value: str): ...

    def removeItem(self, key: str): ...


T = TypeVar('T')


@dataclass
class Restore(Generic[T]):
    clazz: Type[T]
    present: bool
    instance: T | None
    exception: Exception | None = None

    def instance_or_default(self) -> T:
        return self.instance if self.instance else self.clazz()


class State:
    def __init__(self, storage: Storage, key: str):
        self.key = key
        self.storage = storage

    def restore(self, clazz: Type[T]) -> Restore[T]:
        j = self.storage.getItem(self.key)
        if not j:
            return Restore(clazz, False, None)
        try:
            obj = serialization.from_json(j, clazz)
        except Exception as e:
            logger.error(f'failed restore of type {clazz} with json ```{j}```')
            import traceback
            traceback.print_exc()
            return Restore(clazz, True, None, e)
        return Restore(clazz, True, obj)

    def save(self, obj):
        j = serialization.to_json(obj, type(obj))
        self.storage.setItem(self.key, j)


class DictStorage(Storage):

    def __init__(self):
        self.storage = {}

    def keys(self) -> Iterator[str]:
        return iter(self.storage.keys())

    @property
    def length(self) -> int:
        return len(self.storage)

    def items(self) -> Iterator[tuple[str, str]]:
        return iter(self.storage.items())

    def getItem(self, key: str) -> str | None:
        return self.storage.get(key)

    def setItem(self, key: str, value: str):
        self.storage[key] = value

    def removeItem(self, key: str):
        del self.storage[key]


class JsStorage(Storage):
    def __init__(self, js_storage=None):
        if js_storage is None:
            import js
            js_storage = js.window.localStorage
        self.storage = js_storage

    def keys(self) -> Iterator[str]:
        for i in range(self.storage.length):
            yield self.storage.key(i)

    @property
    def length(self) -> int:
        return self.storage.length

    def items(self) -> Iterator[tuple[str, str]]:
        return self.storage.items()

    def getItem(self, key: str) -> str | None:
        return self.storage.getItem(key)

    def setItem(self, key: str, value: str):
        self.storage.setItem(key, value)

    def removeItem(self, key: str):
        self.storage.removeItem(key)



from __future__ import annotations

import logging
import sys
import time
from pathlib import Path

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.common import quickstart, _remote_module_not_found_console
from wwwpy.common.designer import log_emit
from wwwpy.common.rpc.custom_loader import CustomFinder
from wwwpy.common.settingslib import Settings
from wwwpy.resources import library_resources, from_directory
from wwwpy.server import tcp_port
from wwwpy.server.custom_str import CustomStr
from wwwpy.webserver import Webserver
from wwwpy.webservers.available_webservers import available_webservers
from wwwpy.websocket import WebsocketPool

logger = logging.getLogger(__name__)


def start_default(port: int, directory: Path, dev_mode=False, settings: Settings = None):
    webserver = available_webservers().new_instance()

    if quickstart.invalid_project(directory):
        warn_invalid_project(directory)

    convention(directory, webserver, dev_mode=dev_mode, settings=settings)

    while tcp_port.is_port_busy(port):
        logger.warning(f'port {port} is busy, retrying...')
        [time.sleep(0.1) for _ in range(20) if tcp_port.is_port_busy(port)]

    webserver.set_port(port).start_listen()


websocket_pool: WebsocketPool = None


def convention(directory: Path, webserver: Webserver = None, dev_mode=False, settings: Settings = None):
    if settings is None:
        settings = Settings()
    server_rpc_packages = ['server.rpc']
    # todo fix imprecision: for each of the packages, we conceptually apply a remote_proxy_transform
    # we specify also, e.g., 'remote' to transform it but it could end up in an empty string and not a proxy
    # we also should parametrize the transform to make it explicit
    remote_rpc_packages = {'remote', 'remote.rpc', 'wwwpy.remote', 'wwwpy.remote.rpc'}  #
    if dev_mode:
        server_rpc_packages.append('wwwpy.server.designer.rpc')
        remote_rpc_packages.update({'wwwpy.remote.designer', 'wwwpy.remote.designer.rpc'})
        log_emit.add_once(print)
        # quickstart._make_hotreload_work(directory)

    sys.path.insert(0, CustomStr(directory))
    sys.meta_path.insert(0, CustomFinder(remote_rpc_packages))
    global websocket_pool
    websocket_pool = WebsocketPool('/wwwpy/ws')
    services = _configure_server_rpc_services('/wwwpy/rpc', server_rpc_packages)
    services.generate_remote_stubs()
    routes = [services.route, websocket_pool.http_route, *bootstrap_routes(
        resources=[
            library_resources(),
            services.remote_stub_resources(),
            from_directory(directory / 'remote', relative_to=directory),
            from_directory(directory / 'common', relative_to=directory),
        ],
        # language=python
        python=f'from wwwpy.remote.browser_main import entry_point; await entry_point(dev_mode={dev_mode})'
    )]

    if dev_mode:
        import wwwpy.server.designer.dev_mode as dev_modelib
        dev_modelib._warning_on_multiple_clients(websocket_pool)

        dev_modelib.start_hotreload(
            directory, websocket_pool, services,
            server_folders={'common', 'server'},
            remote_folders={'common', 'remote'}
        )
        if settings.hotreload_self:
            logger.info('devself detected')
            import wwwpy
            wwwpy_dir = Path(wwwpy.__file__).parent
            wwwpy_package_dir = wwwpy_dir.parent
            dev_modelib.start_hotreload(
                wwwpy_package_dir, websocket_pool, services,
                server_folders={'wwwpy/common', 'wwwpy/server'},
                remote_folders={'wwwpy/common', 'wwwpy/remote'}
            )

    if webserver is not None:
        webserver.set_http_route(*routes)


from wwwpy.rpc import RpcRoute


def _configure_server_rpc_services(route_path: str, modules: list[str]) -> RpcRoute:
    services = RpcRoute(route_path)

    for module_name in modules:
        try:
            services.add_module(module_name)
        except Exception as e:
            print(f'could not load rpc module `{module_name}`: {e}')
            return None

    return services


def warn_invalid_project(directory: Path):
    content = _remote_module_not_found_console.replace('$[directory]', str(directory.absolute()))
    lines = content.split('\n')
    for line in lines:
        print(line)
    print('Continuing in ', end='', flush=True)
    for text in "3… 2… 1… \n":
        print(text, end='', flush=True)
        time.sleep(0.5)
    print()



from __future__ import annotations

from typing import Callable

import libcst as cst

from wwwpy.common import modlib


def html_string_edit(source_code: str, class_name: str, html_manipulator: Callable[[str], str]) -> str:
    """This function modifies the HTML string in the source code using the provided manipulator function.
    It only modifies the HTML string within the specified class.
    Supports both triple single-quoted and triple double-quoted strings.
    """
    tree = cst.parse_module(source_code)
    transformer = _HTMLStringUpdater(class_name, html_manipulator)
    modified_tree = tree.visit(transformer)
    return modified_tree.code


def html_from(module: str, class_name: str) -> str | None:
    path = modlib._find_module_path(module)
    if not path:
        return None
    source_code = path.read_text()
    return html_from_source(source_code, class_name)


def html_from_source(source_code: str, class_name: str) -> str | None:
    """This function extracts the HTML string from the source code."""
    html_res = []

    def html_manipulator(html):
        html_res.append(html)
        return html

    html_string_edit(source_code, class_name, html_manipulator)  # improper use to capture html
    if not html_res:
        return None
    return html_res[0]


class _HTMLStringUpdater(cst.CSTTransformer):
    def __init__(self, class_name: str, html_manipulator: Callable[[str], str]):
        super().__init__()
        self.class_name = class_name
        self.html_manipulator = html_manipulator
        self.current_class = None

    def visit_ClassDef(self, node: cst.ClassDef) -> None:
        self.current_class = node.name.value

    def leave_ClassDef(self, original_node: cst.ClassDef, updated_node: cst.ClassDef) -> cst.CSTNode:
        self.current_class = None
        return updated_node

    def leave_SimpleString(self, original_node: cst.SimpleString, updated_node: cst.SimpleString) -> cst.CSTNode:
        if self.current_class == self.class_name:
            if (original_node.value.startswith('"""') and original_node.value.endswith('"""')) or \
                    (original_node.value.startswith("'''") and original_node.value.endswith("'''")):
                quote_type = original_node.value[:3]
                original_html = original_node.value[3:-3]
                modified_html = self.html_manipulator(original_html)
                return updated_node.with_changes(value=f'{quote_type}{modified_html}{quote_type}')
        return updated_node



from .. import platformdirs
from ..common.settingslib import Settings


def user_settings() -> Settings:
    _user_config_path = platformdirs.user_config_path('wwwpy', roaming=True, ensure_exists=True)
    settings = Settings()
    settings.load(_user_config_path / 'settings.ini')
    return settings



from pathlib import Path
from typing import List, Any

from wwwpy.common.rpc import serialization
from wwwpy.common.filesystem.sync import Event
from wwwpy.common.filesystem.sync import event_invert_apply


def sync_source(source: Path, events: List[Event]) -> List[Any]:
    events_inverted = event_invert_apply.events_invert(source, events)
    ser = serialization.serialize(events_inverted, List[Event])
    return ser


def sync_target(target_root: Path, changes: List[Any]) -> None:
    events = serialization.deserialize(changes, List[Event])
    event_invert_apply.events_apply(target_root, events)


def sync_init(source: Path) -> List[Any]:
    events = event_invert_apply.events_init(source)
    ser = serialization.serialize(events, List[Event])
    return ser





from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Dict, List, Tuple, Optional

from wwwpy.common.designer.html_parser import html_to_tree, CstNode, CstTree


@dataclass()
class Node:
    """This is intended to be serializable because it could cross the client/server boundary."""
    tag_name: str
    """The tag name in lowercase."""
    child_index: int
    """This is the index in the list of children of the parent node.
    It is -1 if the node has no parent.
    """
    attributes: Dict[str, Optional[str]]
    """The HTML attributes of the node."""

    def __post_init__(self):
        assert self.tag_name == self.tag_name.lower()


IndexPath = List[int]
NodePath = List[Node]
"""This is the path from the root to a node in the DOM tree."""


def path_to_index(path: NodePath) -> IndexPath:
    return [node.child_index for node in path]


def check_node_path(node_path: IndexPath):
    if len(node_path) > 0 and not isinstance(node_path[0], int):
        raise ValueError(f'Invalid node path: {node_path}')


def node_path_serialize(path: NodePath) -> str:
    return json.dumps([node.__dict__ for node in path])


def node_path_deserialize(serialized: str) -> NodePath:
    node_dicts = json.loads(serialized)
    return [Node(**node_dict) for node_dict in node_dicts]


def locate_node(html: str, path: NodePath) -> CstNode | None:
    cst_tree = html_to_tree(html)
    target_node = _locate_node_rec(cst_tree, path)
    return target_node


def _locate_node_rec(nodes: CstTree, path: NodePath, depth: int = 0) -> CstNode | None:
    if depth >= len(path):
        return None

    target_node = path[depth]
    if target_node.child_index < 0 or target_node.child_index >= len(nodes):
        return None
    node = nodes[target_node.child_index]
    if depth == len(path) - 1:
        return node
    return _locate_node_rec(node.children, path, depth + 1)


def locate_node_indexed(html: str, index_path: IndexPath) -> CstNode | None:
    check_node_path(index_path)
    cst_tree = html_to_tree(html)

    def find_node(nodes: CstTree, path: IndexPath, depth: int) -> CstNode | None:
        if depth >= len(path):
            return None

        child_index = path[depth]
        if child_index < 0 or child_index >= len(nodes):
            return None
        node = nodes[child_index]
        if depth == len(path) - 1:
            return node
        return find_node(node.children, path, depth + 1)

    target_node = find_node(cst_tree, index_path, 0)

    return target_node


def locate_span(html: str, path: NodePath) -> Tuple[int, int] | None:
    """This function locates the position of the node specified by the path in the HTML string.
    The position is represented by the start and end indices of the node in the HTML string.
    """

    node = locate_node(html, path)
    return node.span if node else None


def locate_span_indexed(html: str, index_path: IndexPath) -> Tuple[int, int] | None:
    check_node_path(index_path)
    """This function locates the position of the node specified by the path in the HTML string.
    The position is represented by the start and end indices of the node in the HTML string.
    """

    node = locate_node_indexed(html, index_path)
    return node.span if node else None


def tree_to_path(tree: CstTree, index_path: IndexPath) -> NodePath:
    check_node_path(index_path)
    """This function converts a tree of CstNode objects to a NodePath."""

    def _node(index: int, node: CstNode) -> Node:
        return Node(node.tag_name, index, node.attributes)

    result = []
    children = tree
    for index in index_path:
        node = children[index]
        result.append(_node(index, node))
        children = node.children
    return result


def html_to_node_path(html: str, index_path: IndexPath) -> NodePath:
    check_node_path(index_path)
    tree = html_to_tree(html)
    return tree_to_path(tree, index_path)


def rebase_path(source_html: str, live_path: NodePath) -> NodePath:
    source_tree = html_to_tree(source_html)
    cst_node_list = tree_fuzzy_match(source_tree, live_path)
    last_node = cst_node_list[-1]
    result = node_path_from_leaf(last_node)
    return result


def tree_fuzzy_match(tree: CstTree, node_path: NodePath) -> List[CstNode]:
    """This function finds the best CstNode matches in the tree for the given NodePath."""

    def flatten(t: CstTree) -> List[CstNode]:
        result = []
        for node in t:
            result.append(node)
            result.extend(flatten(node.children))
        return result

    nodes = flatten(tree)

    # for all nodes/node_path pair, compute the similarity
    # and store the node with the highest similarity
    matches = []
    for dx in node_path:
        candidates = []
        for sx in nodes:
            similarity = node_similarity(sx, dx)
            candidates.append((similarity, sx))
        candidates.sort(key=lambda pair: pair[0], reverse=True)
        matches.append(candidates)

    # for each node in the node_path, select the best match
    result: NodePath = []
    for match in matches:
        bm = match[0][1]
        result.append(bm)

    return result


def node_similarity(node1: CstNode, node2: CstNode) -> float:
    """This function computes the similarity between two CstNode objects."""
    # assert isinstance(node1, CstNode)
    # assert isinstance(node2, CstNode)
    # import math
    # level_diff = abs(node1.level - node2.level)
    # level_penality = (1.0 - math.log(level_diff + 1)) if level_diff > 0 else 1.0

    if node1.tag_name != node2.tag_name:
        return 0.05
    attr1 = node1.attributes
    attr2 = node2.attributes

    total_keys = set(attr1.keys()) | set(attr2.keys())

    if not total_keys:  # both nodes have no attributes
        return 1.0 # * level_penality

    common_keys = set(attr1.keys()) & set(attr2.keys())

    def attr_eq(key) -> bool:
        return attr1[key] == attr2[key]

    if 'data-name' in common_keys:
        return 1.0 if attr_eq('data-name') else 0.0

    def attr_sim(key) -> float:
        eq_points = 0.25 if attr_eq(key) else -0.25
        return 0.5 + eq_points

    similarity = sum(attr_sim(key) for key in common_keys) / len(total_keys)
    return similarity


def node_path_from_leaf(last_node: CstNode) -> NodePath:
    """This function constructs a NodePath from a leaf node."""
    result = []

    current = last_node
    while current is not None:
        node = Node(current.tag_name, current.child_index, current.attributes)
        result.insert(0, node)
        current = current.parent
    return result



"""At the time of writing this module is used only browser-side,
even though it is in common because of no platform specific dependencies."""
from __future__ import annotations

import inspect
from pathlib import Path


def find_source_file(clazz) -> Path | None:
    """
    Find the source file for a given class.
    E.g., given a [wwwpy.remote.component.Component] it will return the path to the file where the class is defined.
    """
    try:
        file_path = inspect.getfile(clazz)
        path = Path(file_path)
        return path if path.exists() else None
    except TypeError:
        return None
